# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import datetime
from unittest import TestCase

import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import Actor, ActorType, Annotation, AnnotationType
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.annotation_parser import AnnotationParser


def test_parse_annotation():
    annotation_parser = AnnotationParser()
    annotation_dict = {
        "annotationDate": "2010-01-29T18:30:22Z",
        "annotationType": "OTHER",
        "annotator": "Person: Jane Doe ()",
        "comment": "Document level annotation",
    }

    annotation = annotation_parser.parse_annotation(annotation_dict, spdx_id=DOCUMENT_SPDX_ID)

    assert annotation.annotator == Actor(ActorType.PERSON, name="Jane Doe")
    assert annotation.annotation_type == AnnotationType.OTHER
    assert annotation.annotation_date == datetime.datetime(2010, 1, 29, 18, 30, 22)
    assert annotation.annotation_comment == "Document level annotation"
    assert annotation.spdx_id == DOCUMENT_SPDX_ID


def test_parse_all_annotations():
    annotation_parser = AnnotationParser()
    doc_dict = {
        "SPDXID": DOCUMENT_SPDX_ID,
        "packages": [
            {
                "SPDXID": "SPDXRef-Package",
                "annotations": [
                    {
                        "annotationDate": "2010-01-29T17:30:22Z",
                        "annotationType": "REVIEW",
                        "annotator": "Person: Mick Doe ()",
                        "comment": "Package level annotation",
                    }
                ],
            }
        ],
        "files": [
            {
                "SPDXID": "SPDXRef-File",
                "annotations": [
                    {
                        "annotationDate": "2010-01-29T18:30:22Z",
                        "annotationType": "OTHER",
                        "annotator": "Person: Jane Doe ()",
                        "comment": "File level annotation",
                    }
                ],
            }
        ],
        "snippets": [
            {
                "SPDXID": "SPDXRef-Snippet",
                "annotations": [
                    {
                        "annotationDate": "2022-01-29T18:30:32Z",
                        "annotationType": "REVIEW",
                        "annotator": "Person: Jonas Rie (jonas@example.com)",
                        "comment": "Snippet level annotation",
                    }
                ],
            }
        ],
        "revieweds": [
            {"reviewDate": "2010-01-29T18:30:22Z", "reviewer": "Person: Jane Doe ()", "comment": "Review annotation"}
        ],
    }

    annotations = annotation_parser.parse_all_annotations(input_doc_dict=doc_dict)

    assert len(annotations) == 4
    test_case = TestCase()
    test_case.maxDiff = None
    test_case.assertCountEqual(
        annotations,
        [
            Annotation(
                spdx_id=DOCUMENT_SPDX_ID,
                annotation_type=AnnotationType.REVIEW,
                annotator=Actor(actor_type=ActorType.PERSON, name="Jane Doe", email=None),
                annotation_date=datetime.datetime(2010, 1, 29, 18, 30, 22),
                annotation_comment="Review annotation",
            ),
            Annotation(
                spdx_id="SPDXRef-Package",
                annotation_type=AnnotationType.REVIEW,
                annotator=Actor(actor_type=ActorType.PERSON, name="Mick Doe", email=None),
                annotation_date=datetime.datetime(2010, 1, 29, 17, 30, 22),
                annotation_comment="Package level annotation",
            ),
            Annotation(
                spdx_id="SPDXRef-File",
                annotation_type=AnnotationType.OTHER,
                annotator=Actor(actor_type=ActorType.PERSON, name="Jane Doe", email=None),
                annotation_date=datetime.datetime(2010, 1, 29, 18, 30, 22),
                annotation_comment="File level annotation",
            ),
            Annotation(
                spdx_id="SPDXRef-Snippet",
                annotation_type=AnnotationType.REVIEW,
                annotator=Actor(actor_type=ActorType.PERSON, name="Jonas Rie", email="jonas@example.com"),
                annotation_date=datetime.datetime(2022, 1, 29, 18, 30, 32),
                annotation_comment="Snippet level annotation",
            ),
        ],
    )


@pytest.mark.parametrize(
    "incomplete_annotation_dict",
    [
        {"annotator": "Person: Jane Doe ()"},
        {"annotationDate": "2010-01-29T18:30:22Z"},
    ],
)
def test_parse_incomplete_annotation(incomplete_annotation_dict):
    annotation_parser = AnnotationParser()

    with pytest.raises(SPDXParsingError):
        annotation_parser.parse_annotation(incomplete_annotation_dict)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Callable, Dict, List, Optional

from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import raise_parsing_error_if_logger_has_messages


def json_str_to_enum_name(json_str: str) -> str:
    if not isinstance(json_str, str):
        raise SPDXParsingError([f"Type for enum must be str not {type(json_str).__name__}"])
    return json_str.replace("-", "_").upper()


def parse_field_or_log_error(
    logger: Logger,
    field: Any,
    parsing_method: Callable = lambda x: x,
    default: Any = None,
    field_is_list: bool = False,
) -> Any:
    if not field:
        return default
    try:
        if field_is_list:
            return parse_list_of_elements(field, parsing_method)
        else:
            return parsing_method(field)
    except SPDXParsingError as err:
        logger.extend(err.get_messages())
    except (TypeError, ValueError) as err:
        logger.append(err.args[0])
    return default


def append_parsed_field_or_log_error(
    logger: Logger, list_to_append_to: List[Any], field: Any, method_to_parse: Callable
) -> List[Any]:
    try:
        parsed_element = method_to_parse(field)
        list_to_append_to.append(parsed_element)
    except SPDXParsingError as err:
        logger.extend(err.get_messages())
    except (TypeError, ValueError) as err:
        logger.append(err.args[0])
    return list_to_append_to


def parse_field_or_no_assertion_or_none(field: Optional[str], method_for_field: Callable = lambda x: x) -> Any:
    if field == SpdxNoAssertion().__str__():
        return SpdxNoAssertion()
    elif field == SpdxNone().__str__():
        return SpdxNone()
    else:
        return method_for_field(field)


def parse_field_or_no_assertion(field: Optional[str], method_for_field: Callable = lambda x: x) -> Any:
    if field == SpdxNoAssertion().__str__():
        return SpdxNoAssertion()
    else:
        return method_for_field(field)


def parse_list_of_elements(list_of_elements: List[Dict], method_to_parse_element: Callable, logger=None) -> List[Any]:
    if not logger:
        logger = Logger()
    parsed_elements = []
    for element_dict in list_of_elements:
        parsed_elements = append_parsed_field_or_log_error(
            logger, parsed_elements, element_dict, method_to_parse_element
        )
    raise_parsing_error_if_logger_has_messages(logger)
    return parsed_elements


def delete_duplicates_from_list(list_with_potential_duplicates: List[Any]) -> List[Any]:
    list_without_duplicates = list(dict.fromkeys(list_with_potential_duplicates))
    return list_without_duplicates



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from unittest import mock

import pytest
from license_expression import LicenseExpression

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm, Package, PackagePurpose, SpdxNoAssertion, SpdxNone


@mock.patch("spdx_tools.spdx.model.ExternalPackageRef", autospec=True)
@mock.patch("spdx_tools.spdx.model.Checksum", autospec=True)
@mock.patch("spdx_tools.spdx.model.PackageVerificationCode", autospec=True)
@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_correct_initialization(actor, verif_code, checksum, ext_ref):
    package = Package(
        "id",
        "name",
        SpdxNoAssertion(),
        "version",
        "file_name",
        SpdxNoAssertion(),
        actor,
        True,
        verif_code,
        [checksum],
        "homepage",
        "source_info",
        None,
        [spdx_licensing.parse("license and expression"), SpdxNoAssertion()],
        SpdxNone(),
        "comment on license",
        "copyright",
        "summary",
        "description",
        "comment",
        [ext_ref, ext_ref],
        ["text"],
        PackagePurpose.OTHER,
        datetime(2022, 1, 1),
        None,
        None,
    )
    assert package.spdx_id == "id"
    assert package.name == "name"
    assert package.download_location == SpdxNoAssertion()
    assert package.version == "version"
    assert package.file_name == "file_name"
    assert package.supplier == SpdxNoAssertion()
    assert package.originator == actor
    assert package.files_analyzed
    assert package.verification_code == verif_code
    assert package.checksums == [checksum]
    assert package.homepage == "homepage"
    assert package.source_info == "source_info"
    assert package.license_concluded is None
    assert package.license_info_from_files == [spdx_licensing.parse("license and expression"), SpdxNoAssertion()]
    assert package.license_declared == SpdxNone()
    assert package.license_comment == "comment on license"
    assert package.copyright_text == "copyright"
    assert package.summary == "summary"
    assert package.description == "description"
    assert package.comment == "comment"
    assert package.external_references == [ext_ref, ext_ref]
    assert package.attribution_texts == ["text"]
    assert package.primary_package_purpose == PackagePurpose.OTHER
    assert package.release_date == datetime(2022, 1, 1)
    assert package.built_date is None
    assert package.valid_until_date is None


def test_correct_initialization_with_default_values():
    package = Package("id", "name", "location")
    assert package.spdx_id == "id"
    assert package.name == "name"
    assert package.download_location == "location"
    assert package.version is None
    assert package.file_name is None
    assert package.supplier is None
    assert package.originator is None
    assert package.files_analyzed
    assert package.verification_code is None
    assert package.checksums == []
    assert package.homepage is None
    assert package.source_info is None
    assert package.license_concluded is None
    assert package.license_info_from_files == []
    assert package.license_declared is None
    assert package.license_comment is None
    assert package.copyright_text is None
    assert package.summary is None
    assert package.description is None
    assert package.comment is None
    assert package.external_references == []
    assert package.attribution_texts == []
    assert package.primary_package_purpose is None
    assert package.release_date is None
    assert package.built_date is None
    assert package.valid_until_date is None


def test_wrong_type_in_spdx_id():
    with pytest.raises(TypeError):
        Package(42, "name", "location")


def test_wrong_type_in_name():
    with pytest.raises(TypeError):
        Package("id", 42, "location")


def test_wrong_type_in_download_location():
    with pytest.raises(TypeError):
        Package("id", "name", 42)


def test_wrong_type_in_version():
    with pytest.raises(TypeError):
        Package("id", "name", "location", version=42)


def test_wrong_type_in_file_name():
    with pytest.raises(TypeError):
        Package("id", "name", "location", file_name=42)


def test_wrong_type_in_supplier():
    with pytest.raises(TypeError):
        Package("id", "name", "location", supplier=SpdxNone())


def test_wrong_type_in_originator():
    with pytest.raises(TypeError):
        Package("id", "name", "location", originator=SpdxNone())


def test_wrong_type_in_files_analyzed():
    with pytest.raises(TypeError):
        Package("id", "name", "location", files_analyzed=None)


def test_wrong_type_in_verification_code():
    with pytest.raises(TypeError):
        Package("id", "name", "location", verification_code=[])


def test_wrong_type_in_checksums():
    with pytest.raises(TypeError):
        Package("id", "name", "location", checksums=Checksum(ChecksumAlgorithm.MD2, "value"))


def test_wrong_type_in_homepage():
    with pytest.raises(TypeError):
        Package("id", "name", "location", homepage=42)


def test_wrong_type_in_source_info():
    with pytest.raises(TypeError):
        Package("id", "name", "location", source_info=42)


def test_wrong_type_in_license_concluded():
    with pytest.raises(TypeError):
        Package("id", "name", "location", license_concluded=[])


def test_wrong_type_in_license_info_from_files():
    with pytest.raises(TypeError):
        Package("id", "name", "location", license_info_from_files=LicenseExpression("string"))


def test_wrong_type_in_license_declared():
    with pytest.raises(TypeError):
        Package("id", "name", "location", license_declared=[])


def test_wrong_type_in_license_comment():
    with pytest.raises(TypeError):
        Package("id", "name", "location", license_comment=42)


def test_wrong_type_in_copyright_text():
    with pytest.raises(TypeError):
        Package("id", "name", "location", copyright_text=42)


def test_wrong_type_in_summary():
    with pytest.raises(TypeError):
        Package("id", "name", "location", summary=42)


def test_wrong_type_in_description():
    with pytest.raises(TypeError):
        Package("id", "name", "location", description=42)


def test_wrong_type_in_comment():
    with pytest.raises(TypeError):
        Package("id", "name", "location", comment=[])


def test_wrong_type_in_external_references():
    with pytest.raises(TypeError):
        Package("id", "name", "location", external_references=["external_ref"])


def test_wrong_type_in_attribution_texts():
    with pytest.raises(TypeError):
        Package("id", "name", "location", attribution_texts="text")


def test_wrong_type_in_primary_package_purpose():
    with pytest.raises(TypeError):
        Package("id", "name", "location", primary_package_purpose=[])


def test_wrong_type_in_release_date():
    with pytest.raises(TypeError):
        Package("id", "name", "location", release_date=42)


def test_wrong_type_in_built_date():
    with pytest.raises(TypeError):
        Package("id", "name", "location", built_date="2022-01-01")


def test_wrong_type_in_valid_until_date():
    with pytest.raises(TypeError):
        Package("id", "name", "location", valid_until_date=SpdxNone())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm, File, PackageVerificationCode
from spdx_tools.spdx.spdx_element_utils import calculate_file_checksum, calculate_package_verification_code


@pytest.fixture
def generate_test_files(tmp_path):
    file_path_1 = tmp_path.joinpath("file1")
    file_path_2 = tmp_path.joinpath("file2")

    with open(file_path_1, "wb") as file:
        file.write(bytes(111))
    with open(file_path_2, "wb") as file:
        file.write(bytes(222))

    yield str(file_path_1), str(file_path_2)


def test_file_checksum_calculation(generate_test_files):
    filepath1, filepath2 = generate_test_files
    checksum = calculate_file_checksum(filepath1, ChecksumAlgorithm.SHA1)
    assert checksum == "dd90903d2f566a3922979dd5e18378a075c7ed33"
    checksum = calculate_file_checksum(filepath2, ChecksumAlgorithm.SHA1)
    assert checksum == "140dc52658e2eeee3fdc4d471cce84fec7253fe3"


def test_verification_code_calculation_with_predefined_checksums(generate_test_files):
    filepath1, filepath2 = generate_test_files
    file1 = File(
        filepath1,
        "SPDXRef-hello",
        [Checksum(ChecksumAlgorithm.SHA1, "20862a6d08391d07d09344029533ec644fac6b21")],
    )
    file2 = File(
        filepath2,
        "SPDXRef-Makefile",
        [Checksum(ChecksumAlgorithm.SHA1, "69a2e85696fff1865c3f0686d6c3824b59915c80")],
    )
    verification_code = calculate_package_verification_code([file1, file2])

    assert verification_code == PackageVerificationCode("c6cb0949d7cd7439fce8690262a0946374824639")


def test_verification_code_calculation_with_calculated_checksums(generate_test_files):
    filepath1, filepath2 = generate_test_files
    file1 = File(
        filepath1,
        "SPDXRef-hello",
        [Checksum(ChecksumAlgorithm.MD4, "20862a6d08391d07d09344029533ec644fac6b21")],
    )
    file2 = File(
        filepath2,
        "SPDXRef-Makefile",
        [Checksum(ChecksumAlgorithm.MD4, "69a2e85696fff1865c3f0686d6c3824b59915c80")],
    )
    verification_code = calculate_package_verification_code([file1, file2])

    assert verification_code == PackageVerificationCode("6f29d813abb63ee52a47dbcb691ea2e70f956328")


def test_verification_code_calculation_with_wrong_file_location():
    unknown_file_name = "./unknown_file_name"
    file1 = File(
        unknown_file_name,
        "SPDXRef-unknown",
        [Checksum(ChecksumAlgorithm.MD4, "20862a6d08391d07d09344029533ec644fac6b21")],
    )

    with pytest.raises(FileNotFoundError) as err:
        calculate_package_verification_code([file1])

    assert unknown_file_name in str(err.value)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import field
from datetime import datetime
from enum import Enum, auto

from beartype.typing import Dict, List, Optional, Union
from license_expression import LicenseExpression

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx.model import Actor, Checksum, SpdxNoAssertion, SpdxNone


class PackagePurpose(Enum):
    APPLICATION = auto()
    FRAMEWORK = auto()
    LIBRARY = auto()
    CONTAINER = auto()
    OPERATING_SYSTEM = auto()
    DEVICE = auto()
    FIRMWARE = auto()
    SOURCE = auto()
    ARCHIVE = auto()
    FILE = auto()
    INSTALL = auto()
    OTHER = auto()


@dataclass_with_properties
class PackageVerificationCode:
    value: str
    excluded_files: List[str] = field(default_factory=list)

    def __init__(self, value: str, excluded_files: List[str] = None):
        excluded_files = [] if excluded_files is None else excluded_files
        check_types_and_set_values(self, locals())


class ExternalPackageRefCategory(Enum):
    SECURITY = auto()
    PACKAGE_MANAGER = auto()
    PERSISTENT_ID = auto()
    OTHER = auto()


CATEGORY_TO_EXTERNAL_PACKAGE_REF_TYPES: Dict[ExternalPackageRefCategory, List[str]] = {
    ExternalPackageRefCategory.SECURITY: ["cpe22Type", "cpe23Type", "advisory", "fix", "url", "swid"],
    ExternalPackageRefCategory.PACKAGE_MANAGER: ["maven-central", "npm", "nuget", "bower", "purl"],
    ExternalPackageRefCategory.PERSISTENT_ID: ["swh", "gitoid"],
    ExternalPackageRefCategory.OTHER: [],
}


@dataclass_with_properties
class ExternalPackageRef:
    category: ExternalPackageRefCategory
    # In theory, once could refine the typing,
    # see https://spdx.github.io/spdx-spec/v2.3/external-repository-identifiers/. But it's probably not worth the
    # effort.
    reference_type: str
    locator: str
    comment: Optional[str] = None

    def __init__(
        self, category: ExternalPackageRefCategory, reference_type: str, locator: str, comment: Optional[str] = None
    ):
        check_types_and_set_values(self, locals())


@dataclass_with_properties
class Package:
    spdx_id: str
    name: str
    download_location: Union[str, SpdxNoAssertion, SpdxNone]
    version: Optional[str] = None
    file_name: Optional[str] = None
    supplier: Optional[Union[Actor, SpdxNoAssertion]] = None
    originator: Optional[Union[Actor, SpdxNoAssertion]] = None
    files_analyzed: bool = True
    verification_code: Optional[PackageVerificationCode] = None
    checksums: List[Checksum] = field(default_factory=list)
    homepage: Optional[Union[str, SpdxNoAssertion, SpdxNone]] = None
    source_info: Optional[str] = None
    license_concluded: Optional[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = None
    license_info_from_files: List[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = field(default_factory=list)
    license_declared: Optional[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = None
    license_comment: Optional[str] = None
    copyright_text: Optional[Union[str, SpdxNoAssertion, SpdxNone]] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    comment: Optional[str] = None
    external_references: List[ExternalPackageRef] = field(default_factory=list)
    attribution_texts: List[str] = field(default_factory=list)
    primary_package_purpose: Optional[PackagePurpose] = None
    release_date: Optional[datetime] = None
    built_date: Optional[datetime] = None
    valid_until_date: Optional[datetime] = None

    def __init__(
        self,
        spdx_id: str,
        name: str,
        download_location: Union[str, SpdxNoAssertion, SpdxNone],
        version: Optional[str] = None,
        file_name: Optional[str] = None,
        supplier: Optional[Union[Actor, SpdxNoAssertion]] = None,
        originator: Optional[Union[Actor, SpdxNoAssertion]] = None,
        files_analyzed: bool = True,
        verification_code: Optional[PackageVerificationCode] = None,
        checksums: List[Checksum] = None,
        homepage: Optional[Union[str, SpdxNoAssertion, SpdxNone]] = None,
        source_info: Optional[str] = None,
        license_concluded: Optional[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = None,
        license_info_from_files: List[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = None,
        license_declared: Optional[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = None,
        license_comment: Optional[str] = None,
        copyright_text: Optional[Union[str, SpdxNoAssertion, SpdxNone]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        comment: Optional[str] = None,
        external_references: List[ExternalPackageRef] = None,
        attribution_texts: List[str] = None,
        primary_package_purpose: Optional[PackagePurpose] = None,
        release_date: Optional[datetime] = None,
        built_date: Optional[datetime] = None,
        valid_until_date: Optional[datetime] = None,
    ):
        checksums = [] if checksums is None else checksums
        license_info_from_files = [] if license_info_from_files is None else license_info_from_files
        external_references = [] if external_references is None else external_references
        attribution_texts = [] if attribution_texts is None else attribution_texts
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List

from spdx_tools.spdx.model import Document, RelationshipType
from spdx_tools.spdx.model.relationship_filters import filter_by_type_and_origin, filter_by_type_and_target
from spdx_tools.spdx.validation.annotation_validator import validate_annotations
from spdx_tools.spdx.validation.creation_info_validator import validate_creation_info
from spdx_tools.spdx.validation.extracted_licensing_info_validator import validate_extracted_licensing_infos
from spdx_tools.spdx.validation.file_validator import validate_files
from spdx_tools.spdx.validation.package_validator import validate_packages
from spdx_tools.spdx.validation.relationship_validator import validate_relationships
from spdx_tools.spdx.validation.snippet_validator import validate_snippets
from spdx_tools.spdx.validation.spdx_id_validators import get_list_of_all_spdx_ids
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


def validate_full_spdx_document(document: Document, spdx_version: str = None) -> List[ValidationMessage]:
    validation_messages: List[ValidationMessage] = []

    # SPDX version validation has to happen here because subsequent validators rely on it
    document_version: str = document.creation_info.spdx_version
    context = ValidationContext(spdx_id=document.creation_info.spdx_id, element_type=SpdxElementType.DOCUMENT)
    if not spdx_version:
        spdx_version = document_version

    if document_version not in ["SPDX-2.2", "SPDX-2.3"]:
        validation_messages.append(
            ValidationMessage(
                f'only SPDX versions "SPDX-2.2" and "SPDX-2.3" are supported, but the document\'s spdx_version is: '
                f"{document_version}",
                context,
            )
        )
    elif spdx_version != document_version:
        validation_messages.append(
            ValidationMessage(
                f"provided SPDX version {spdx_version} does not match "
                f"the document's SPDX version {document_version}",
                context,
            )
        )

    if validation_messages:
        validation_messages.append(
            ValidationMessage(
                "There are issues concerning the SPDX version of the document. "
                "As subsequent validation relies on the correct version, "
                "the validation process has been cancelled.",
                context,
            )
        )
        return validation_messages

    validation_messages.extend(validate_creation_info(document.creation_info, spdx_version))
    validation_messages.extend(validate_packages(document.packages, spdx_version, document))
    validation_messages.extend(validate_files(document.files, spdx_version, document))
    validation_messages.extend(validate_snippets(document.snippets, spdx_version, document))
    validation_messages.extend(validate_annotations(document.annotations, document))
    validation_messages.extend(validate_relationships(document.relationships, spdx_version, document))
    validation_messages.extend(validate_extracted_licensing_infos(document.extracted_licensing_info))

    document_id = document.creation_info.spdx_id
    document_describes_relationships = filter_by_type_and_origin(
        document.relationships, RelationshipType.DESCRIBES, document_id
    )
    described_by_document_relationships = filter_by_type_and_target(
        document.relationships, RelationshipType.DESCRIBED_BY, document_id
    )

    only_a_single_package = len(document.packages) == 1 and not document.files and not document.snippets
    if not only_a_single_package and not document_describes_relationships + described_by_document_relationships:
        validation_messages.append(
            ValidationMessage(
                f'there must be at least one relationship "{document_id} DESCRIBES ..." or "... DESCRIBED_BY '
                f'{document_id}" when there is not only a single package present',
                ValidationContext(spdx_id=document_id, element_type=SpdxElementType.DOCUMENT),
            )
        )

    all_spdx_ids: List[str] = get_list_of_all_spdx_ids(document)
    auxiliary_set = set()
    duplicated_spdx_ids = set(
        spdx_id for spdx_id in all_spdx_ids if spdx_id in auxiliary_set or auxiliary_set.add(spdx_id)
    )

    if duplicated_spdx_ids:
        validation_messages.append(
            ValidationMessage(
                f"every spdx_id must be unique within the document, but found the following duplicates: "
                f"{sorted(duplicated_spdx_ids)}",
                context,
            )
        )

    return validation_messages
